
# Data Manipulation

from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from Bio.Alphabet import generic_dna
import numpy as np

nucnames = ['A','G','C','T']

def loadData(filename,data_type):
    data = []
    id_seq = []
    for seq_record in SeqIO.parse(filename,data_type):
        id_seq.append(seq_record.id)
        data.append(list(seq_record.seq))

    return data, id_seq

def saveData(data, id_seq, filename, data_type):
    my_seq = []
    for i,seq in enumerate(data):
        my_seq.append(SeqRecord(Seq(''.join(seq),generic_dna),id=str(id_seq[i]),description=''))
    
    with open(filename,"w") as output_file:
        SeqIO.write(my_seq, output_file, data_type)

# sampler for a discrete distribution
def statSamp(pden, idx=False):
    cumsum = pden.cumsum()
    u = np.random.uniform()
    for i in range(4):
        if u <= cumsum[i]:
            if idx:
                return i
            else:
                return nucnames[i]

# simulate taxon sequences given the tree topology and rate matrices
def treeSimu(tree, D, U, U_inv, pden, ndata, missrate=0.0):
    ntips = len(tree)
    pt_matrix = [np.zeros((4,4)) for i in range(2*ntips-3)]
    
    # run tree traversal to acquire the transition matrices
    for node in tree.traverse("postorder"):
        if not node.is_root():
            # pt_matrix[node.name] = np.transpose(np.linalg.lstsq(U.T,
            #                     np.dot(np.diag(np.exp(D*node.dist)),U.T))[0])
            pt_matrix[node.name] = np.dot(U, np.dot(np.diag(np.exp(D*node.dist)), U_inv))
    
    simuData = []
    status = [''] * (2*ntips-2) 
    for run in range(ndata):
        for node in tree.traverse("preorder"):
            if node.is_root():
                status[node.name] = statSamp(pden,idx=True)
            else:
                status[node.name] = statSamp(pt_matrix[node.name][status[node.up.name]],idx=True)
            
        simuData.append([nucnames[i] if np.random.uniform() > missrate else '-' for i in status[:ntips]])
    
    return np.transpose(simuData)
    


import numpy as np
import re
import os
from Bio import Phylo
from cStringIO import StringIO
from .treeManipulation import init, namenum
from phyloinfer import Tree
from collections import defaultdict

try:
    from pyTQDist import quartetDistance as qdist
except ImportError:
    print "No module named pyTQDist found!"



def treeStats(sampled_tree, target_tree_dict=None):
    if not target_tree_dict:
        id2stats = defaultdict(int)
        id2tree = defaultdict()
        for tree in sampled_tree:
            ID = tree.get_topology_id()
            id2stats[ID] += 1
            if ID not in id2tree:
                id2tree[ID] = tree

    else:
        id2stats = {tree.get_topology_id():0.0 for _,tree in target_tree_dict.items()}
        id2tree = {tree.get_topology_id():_ for _,tree in target_tree_dict.items()}
        for tree in sampled_tree:
            ID = tree.get_topology_id()
            if ID in id2stats:
                id2stats[ID] += 1     

    tree_samp_count = len(sampled_tree)*1.0
    for tree_id in id2stats:
        id2stats[tree_id] /= tree_samp_count
    return sorted(id2stats.items(), key=lambda x:x[1], reverse=True), id2tree


def savePara(sampled_branch, filename):
    import uuid
    branch_count = len(sampled_branch[0])
    with open(filename,'w') as output_file:
        output_file.write('[ID:{}]\n'.format(uuid.uuid4()))
        output_file.write('nIter' + '\t' + '\t'.join(['length[{}]'.format(i) for i in range(branch_count)]) + '\n')
        for i, branch in enumerate(sampled_branch):
            output_file.write(str(i) + '\t' + '\t'.join([str(branch) for branch in sampled_branch[i]]) + '\n')

            
def readPara(filename):
    with open(filename,'r') as readin_file:
        id_line = readin_file.readline()
        ID = ''.join(re.split('\[|\]|ID:',id_line)).strip()
        name_line = readin_file.readline()
        names = name_line.strip('\n').split('\t')
        stats_dict = {name:[] for name in names}
        
        while True:
            line = readin_file.readline()
            if line == "": 
                break
            for i, stats in enumerate(line.strip('\n').split('\t')):
                stats_dict[names[i]].append(float(stats))
            
    return stats_dict, names, ID 


def saveTree(sampled_tree, filename, tree_format, withid=True):
    if type(sampled_tree) is not list:
        sampled_tree = [sampled_tree]
       
    with open(filename,'w') as output_file:
        if withid:
            import uuid
            output_file.write('[ID:{}]\n'.format(uuid.uuid4()))
        for i, tree in enumerate(sampled_tree):
            tree_newick = tree.write(format=tree_format)
            if withid:
                output_file.write('tree_{}'.format(i) + '\t' + tree_newick + '\n')
            else:
                output_file.write(tree_newick + '\n')
            

def readTree(filename, tree_format=3):
    with open(filename,'r') as readin_file:
        id_line = readin_file.readline()
        ID = ''.join(re.split('\[|\]|ID:',id_line)).strip()
        samp_tree_list = []
        while True:
            line = readin_file.readline()
            if line == "":
                break
            tree_name, newick = line.strip('\n').split('\t')
            samp_tree_list.append(Tree(newick,format=tree_format))
            init(samp_tree_list[-1], name='interior')
        
    return samp_tree_list
    

def Qdist(tree1, tree2):
    if not os.path.exists('out'):
        os.makedirs('out')
        
    saveTree(tree1, 'out/tree1.new', 9)
    saveTree(tree2, 'out/tree2.new', 9)
    
    return qdist('out/tree1.new', 'out/tree2.new')


def mcmc_treeprob(filename, data_type, truncate=10, taxon=None):
    mcmc_samp_tree_stats = Phylo.parse(filename, data_type)
    mcmc_samp_tree_dict = {}
    mcmc_samp_tree_name = []
    mcmc_samp_tree_wts = []
    num_hp_tree = 0
    for tree in mcmc_samp_tree_stats:
        handle = StringIO()
        Phylo.write(tree, handle,'newick')
        mcmc_samp_tree_dict[tree.name] = Tree(handle.getvalue().strip())
        if taxon:
            namenum(mcmc_samp_tree_dict[tree.name], taxon)
        else:
            init(mcmc_samp_tree_dict[tree.name],name='interior')
            
        handle.close()
        mcmc_samp_tree_name.append(tree.name)
        mcmc_samp_tree_wts.append(tree.weight)
        num_hp_tree += 1
        if num_hp_tree >= truncate: break
    
    return mcmc_samp_tree_dict, mcmc_samp_tree_name, mcmc_samp_tree_wts
    
        


# Hamiltonian Monte Carlo 

import copy
import sys
import numpy as np
import random
from .rateMatrix import decompHKY, decompJC, decompGTR
from .treeManipulation import idx2nodeMAP, NNI
from .Loglikelihood import phyloLoglikelihood, initialCLV 
from .Logposterior import Logpost, GradLogpost
from .refLeapfrog import reflection, refraction
from .branchManipulation import set



def hmc_iter(curr_tree, curr_branch, curr_U, D, U, U_inv, pden, L, nLeap, stepsz, 
             scale=0.1, delta=0.01, randomization=True, surrogate=False, include=False,
             monitor_event=False):
    # initialize the momentum
    propM = np.random.normal(size=len(curr_branch))
    currM = propM
    currH = curr_U + 0.5 * sum(currM*currM)
    propB = curr_branch
    
    # make a deep copy of the current tree
    prop_tree = copy.deepcopy(curr_tree)
    idx2node = idx2nodeMAP(prop_tree)
    
    if randomization:
        nLeap_exact = np.random.randint(1,nLeap+1)
    else:
        nLeap_exact = nLeap
    
    NNI_attempts, Ref_attempts = 0, 0
    # run the refractive leap-frog
    for i in range(nLeap_exact):
        propM = propM - stepsz/2 * GradLogpost(prop_tree, propB, D, U, U_inv, pden, L,
                                               scale=scale, delta=delta, surrogate=surrogate) 
        
        prop_tree, tmpB, propM, step_nni_attempts, step_ref_attempts = refraction(prop_tree, propB, propM, D, U, U_inv, pden, L,
                                                                                  idx2node, stepsz, delta=delta, scale=scale,
                                                                                  surrogate=surrogate, include=include)
        
        NNI_attempts += step_nni_attempts
        Ref_attempts += step_ref_attempts

        propB = tmpB     
        set(prop_tree,propB)
            
        propM = propM - stepsz/2 * GradLogpost(prop_tree, propB, D, U, U_inv, pden, L,
                                               scale=scale, delta=delta, surrogate=surrogate) 
    
    prop_U = Logpost(prop_tree, propB, D, U, U_inv, pden, L, scale)
    propH = prop_U + 0.5 * sum(propM*propM)
    
    if monitor_event:
        print "NNI attempts: {}\nRef attempts: {}".format(NNI_attempts, Ref_attempts)
        sys.stdout.flush()
    
    ratio = currH - propH
    if ratio >= min(0,np.log(np.random.uniform())):
        return prop_tree, propB, prop_U, NNI_attempts, 1, min(1.0,np.exp(ratio))
    else:
        return curr_tree, curr_branch, curr_U, NNI_attempts, 0, min(1.0,np.exp(ratio))
    

def hmc(curr_tree, curr_branch, Qmat_para, data, nLeap, stepsz, nIter, subModel='HKY', randomization=True,
        surrogate=False, burnin_frac=0.5, adap_stepsz_rate=0.4, scale=0.1, delta=0.01, include=False,
        printfreq=100, monitor_event=False, output_filename=None):
    sampled_tree = []
    sampled_branch = []
    path_U = []
    path_Loglikelihood = []
    nni_attempts = []
    accept_rate = []
    accept_count = 0.0
    burnin = np.floor(burnin_frac*nIter)
        
    # eigen decomposition of the rate matrix
    if subModel == 'JC':
        pden, _ = Qmat_para
        D, U, U_inv, _ = decompJC()
    else:
        pden, sub_rate = Qmat_para
        if subModel == 'HKY':
            D, U, U_inv, _ = decompHKY(pden, sub_rate)
        if subModel == 'GTR':
            AG, AC, AT, GC, GT, CT = sub_rate
            D, U, U_inv, _ = decompGTR(pden, AG, AC, AT, GC, GT, CT)
    
    # initialize the conditional likelihood on tips
    if len(curr_tree) != len(data):
        print "#tips do not match the data!!!"
        return
    L = initialCLV(data)
    
    curr_U = Logpost(curr_tree, curr_branch, D, U, U_inv, pden, L)
    path_U.append(curr_U)
    path_Loglikelihood.append(phyloLoglikelihood(curr_tree, curr_branch, D, U, U_inv, pden, L))
    
    # save current samples to files
    if output_filename:
        import uuid
        samp_tree_file = open(output_filename + '.tree','w')
        samp_para_file = open(output_filename + '.para','w')
        samp_stat_file = open(output_filename + '.stat','w')
        ID = uuid.uuid4()
        
        samp_tree_file.write('[ID:{}]\n'.format(ID))
        samp_tree_file.write('tree_0:' + '\t' + curr_tree.write(format=3) + '\n')
        samp_para_file.write('[ID:{}]\n'.format(ID))
        samp_para_file.write('nIter\t' + 'LnL\t' + '\t'.join(['length[{}]'.format(i) for i in range(len(curr_branch))]) + '\n')
        samp_para_file.write('0\t' + '{}\t'.format(path_Loglikelihood[-1]) 
                             + '\t'.join([str(branch) for branch in curr_branch]) + '\n')
        samp_stat_file.write('[ID:{}]\n'.format(ID))
        samp_stat_file.write('nIter\t' + 'AP\t' + 'NNI\n')
        samp_tree_file.flush()
        samp_para_file.flush()
        samp_stat_file.flush()
        
    Accepted = 0.0
    for i in range(nIter):
        exact_stepsz = np.power(1-adap_stepsz_rate,max(1-i*1.0/burnin,0))*stepsz        
        curr_tree, curr_branch, curr_U, NNI_attempts, accepted, ar = hmc_iter(curr_tree, curr_branch, curr_U,
                                                                D, U, U_inv, pden, L, nLeap, exact_stepsz,
                                                                scale, delta, randomization, surrogate, include,
                                                                monitor_event)
        
        path_U.append(curr_U)
        path_Loglikelihood.append(phyloLoglikelihood(curr_tree, curr_branch, D, U, U_inv, pden, L))
        if output_filename:
            samp_tree_file.write('tree_{}:'.format(i+1) + '\t' + curr_tree.write(format=3) + '\n')
            samp_para_file.write('{}\t{}\t'.format(i+1,path_Loglikelihood[-1]) + 
                                                   '\t'.join([str(branch) for branch in curr_branch]) + '\n')
            samp_stat_file.write('{}\t{}\t{}\n'.format(i+1,ar,NNI_attempts))
            samp_tree_file.flush()
            samp_para_file.flush()
            samp_stat_file.flush()
            
        accept_rate.append(ar)
        nni_attempts.append(NNI_attempts)
        print "{} iteration: current Loglikelihood = {}".format(i+1,path_Loglikelihood[-1])
        sys.stdout.flush()
        if i>=burnin:
            sampled_tree.append(curr_tree)
            sampled_branch.append(curr_branch)
            accept_count += accepted
        Accepted += accepted
        if (i+1)%printfreq == 0:
            print "\n######   {} iterations completed; acceptance rate: {}\n".format(i+1, Accepted/printfreq)
            sys.stdout.flush()
            Accepted = 0.0
            
    if output_filename:
        samp_tree_file.close()
        samp_para_file.close()
        samp_stat_file.close()
        
    print "Overall acceptance rate: {}".format(accept_count/(nIter-burnin))
    return sampled_tree, sampled_branch, path_U, path_Loglikelihood, nni_attempts, accept_rate, accept_count 

    
   


import numpy as np
# tree branch length manipulations

def update(tree, dbranch):
    for node in tree.traverse("postorder"):
        if not node.is_root():
            node.dist += dbranch[node.name]

def get(tree):
    # branch = [0] * (2*len(tree)-3)
    branch = np.zeros(2*len(tree)-3)
    for node in tree.traverse("postorder"):
        if not node.is_root():
            branch[node.name] = node.dist
    return branch

def set(tree,branch):
    for node in tree.traverse("postorder"):
        if not node.is_root():
            node.dist = branch[node.name]


import numpy as np
from phyloinfer import Tree



# Tree initialization: set the node name to the sequence order
def init(tree, branch=None, name='all', scale=0.1, display=False, return_map=False):
    if return_map: idx2node = {}
    i, j = 0, len(tree)
    for node in tree.traverse("postorder"):
        if node.is_leaf():
            if name != 'interior':
                node.name, i = i, i+1
            else:
                node.name = int(node.name)
        else:
            node.name, j = j, j+1
        if not node.is_root():
            if isinstance(branch, basestring) and branch=='random':
                node.dist = np.random.exponential(scale)
            elif branch is not None:
                node.dist = branch[node.name]
        else:
            node.dist = 0.0
            
        if return_map: idx2node[node.name] = node
        if display:
            print node.name, node.dist
        
    if return_map: return idx2node
    
    
def create(ntips, branch='random', scale=0.1):
    tree = Tree()
    tree.populate(ntips)
    tree.unroot()
    init(tree, branch=branch, scale=scale)
    
    return tree
    

def namenum(tree, taxon, nodetosplitMap=None):
    taxon2idx = {}
    j = len(taxon)
    if nodetosplitMap:
        idx2split = ['']*(2*j-3)
    for i, name in enumerate(taxon):
        taxon2idx[name] = i
    for node in tree.traverse("postorder"):
        if node.is_leaf():
            # assert type(node.name) is str, "The taxon name should be strings"
            if not isinstance(node.name, str):
                warnings.warn("The taxon names are not strings, please check if they are already integers!")
            else:
                node.name = taxon2idx[node.name]
                if nodetosplitMap:
                    idx2split[node.name] = nodetosplitMap[node]
        else:
            node.name, j = j, j+1
            if nodetosplitMap and not node.is_root():
                idx2split[node.name] = nodetosplitMap[node]
    
    if nodetosplitMap:
        return idx2split

    

def nametaxon(tree, taxon):
    for node in tree.traverse("postorder"):
        if node.is_leaf():
            assert type(node.name) is int, "The taxon name should be integers"
            node.name = taxon[node.name]            


def is_internal(node):
    return not node.is_leaf() and not node.is_root()


# obtain a dictionary of nodes
def idx2nodeMAP(tree):
    idx2node = {}
    for node in tree.traverse("postorder"):
        idx2node[node.name] = node
    return idx2node


# perform NNI on interior branches
def NNI(node, include=False):
    if node.is_root() or node.is_leaf():
        print "Can not perform NNI on root or leaf branches!"
    else:
        if include:
            neighboor = np.random.randint(3)
            xchild = node.get_sisters()[0]
            parent = node.up
            if neighboor == 1:
                ychild = node.children[0]
            if neighboor == 2:
                ychild = node.children[1]
            if neighboor:
                xchild.detach()
                ychild.detach()
                parent.add_child(ychild)
                node.add_child(xchild)
                return 1
        else:
            neighboor = np.random.randint(2)
            xchild = node.get_sisters()[0]
            parent = node.up
            ychild = node.children[neighboor]
            xchild.detach()
            ychild.detach()
            parent.add_child(ychild)
            node.add_child(xchild)
            return 1
    
    return 0

                        
def copy(tree, branch=None):
    copy_tree = tree.copy('newick')
    for node in copy_tree.traverse('postorder'):
        node.name = int(node.name)
        if not node.is_root() and branch:
            node.dist = branch[node.name]
                
    return copy_tree                    


# Compute the log-posterior function

from .Loglikelihood import phyloLoglikelihood
from .Logprior import phyloLogprior_exp
import numpy as np



# a mollifier for surrogate construction
def mollifier(x, delta):
    if x >= delta:
        return x
    else:
        return 1.0/2/delta *(x*x+delta*delta)
    
# negative log-posterior
def Logpost(tree, branch, D, U, U_inv, pden, L, scale=0.1, surrogate=False, delta = 0.01):
    if surrogate:
        branch = [mollifier(blen,delta) for blen in branch]
        
    return -phyloLoglikelihood(tree, branch, D, U, U_inv, pden, L) - phyloLogprior_exp(branch, scale)


def GradLogpost(tree, branch, D, U, U_inv, pden, L, scale=0.1, surrogate=False, delta = 0.01):
    if surrogate:
        maped_branch = [mollifier(blen,delta) for blen in branch]
        return (-phyloLoglikelihood(tree, maped_branch, D, U, U_inv, pden, L, grad=True)  
                - phyloLogprior_exp(maped_branch, scale, grad=True)) * np.minimum(branch,delta)/delta
    else:    
        return -phyloLoglikelihood(tree, branch, D, U, U_inv, pden, L, grad=True) - phyloLogprior_exp(branch, scale, grad=True)



# Loglikelihood and its gradients
import numpy as np
# from warnings import warn
import warnings
warnings.simplefilter('always', UserWarning)
nuc2vec = {'A':[1.,0.,0.,0.], 'G':[0.,1.,0.,0.], 'C':[0.,0.,1.,0.], 'T':[0.,0.,0.,1.],
           '-':[1.,1.,1.,1.], '?':[1.,1.,1.,1.], 'N':[1.,1.,1.,1.], 'R':[1.,1.,0.,0.],
           'Y':[0.,0.,1.,1.], 'S':[0.,1.,1.,0.], 'W':[1.,0.,0.,1.], 'K':[0.,1.,0.,1.],
           'M':[1.,0.,1.,0.], 'B':[0.,1.,1.,1.], 'D':[1.,1.,0.,1.], 'H':[1.,0.,1.,1.],
           'V':[1.,1.,1.,0.], '.':[1.,1.,1.,1.], 'U':[0.,0.,0.,1.]}


# initialize the conditional likelihood vectors, added unique site option
def initialCLV(data, unique_site=False):
    nseqs, nsites = len(data), len(data[0])
    if unique_site:
        data_arr = np.array(zip(*data))
        unique_sites, counts = np.unique(data_arr, return_counts=True, axis=0)
        n_unique_sites = len(counts)
        unique_data = unique_sites.T
        
        L = np.ones((2*nseqs-2, 4, n_unique_sites))
        for i in range(nseqs):
            L[i] = np.transpose([nuc2vec[c] for c in unique_data[i]])
        return L, counts
    else:   
        L = np.ones((2*nseqs-2, 4, nsites))
        for i in range(nseqs):
            L[i] = np.transpose([nuc2vec[c] for c in data[i]])
        return L

            
def phyloLoglikelihood(tree, branch, D, U, U_inv, pden, L, site_counts=1.0, grad=False, value_and_grad=False):
    nsites = L[0].shape[1]

    Loglikelihood = 0 
    GradLoglikelihood = np.zeros(len(branch))    
    Down = [1.0] * len(branch)
    pt_matrix = [0.0] * len(branch)
    
    for node in tree.traverse("postorder"):
        if not node.is_leaf():
            L[node.name] = 1.0
            for child in node.children:
                pt_matrix[child.name] = np.dot(U * np.exp(D*branch[child.name]), U_inv)
                Down[child.name] = np.dot(pt_matrix[child.name],L[child.name])
                L[node.name] *= Down[child.name] 
            scaler = np.sum(L[node.name],axis=0)
            if not np.all(scaler>0):
                warnings.warn("Caution! Incompatible data detected!")
                return -np.inf
                
            L[node.name] /= scaler
            Loglikelihood += np.sum(np.log(scaler) * site_counts)
            
    Loglikelihood += np.sum(np.log(np.dot(pden,L[node.name])) * site_counts)
    
    if not grad:
        return Loglikelihood
    
    Up = [1.0] * (len(branch)+1)
    for node in tree.traverse("preorder"):
        if node.is_root():
            Up[node.name] = np.repeat(pden.reshape(-1,1), nsites, axis=1)
        else:
            for sister in node.get_sisters():
                Up[node.name] *= Down[sister.name]
            Up[node.name] *= Up[node.up.name]
            pt_matrix_grad = np.dot(U * (D*np.exp(D*branch[node.name])), U_inv)
            
            gradient = np.sum(Up[node.name] * np.dot(pt_matrix_grad,L[node.name]),axis=0)
            
            Up[node.name] = np.dot(pt_matrix[node.name].T,Up[node.name])
            gradient /= np.sum(Up[node.name]* L[node.name],axis=0)
            GradLoglikelihood[node.name] = np.sum(gradient * site_counts)
            
            if not node.is_leaf():
                scaler = np.sum(Up[node.name],axis=0)
                Up[node.name] /= scaler  
    
    if not value_and_grad:            
        return GradLoglikelihood
    else:
        return Loglikelihood, GradLoglikelihood


# compute the log marginal likelihood of a tree given conjugate Gamma prior for branch lengths
def phyloLogMarginallikelihood(tree, D, U, pden, L, gamma_prior=(1,10)):
    nsites = L[0].shape[1]
    ntips = len(tree)
    marginalLoglikelihood = 0    
    Down = [np.ones((4,nsites)) for i in range(2*ntips-3)]
    marginal_pt_matrix = [np.zeros((4,4)) for i in range(2*ntips-3)]
    k, theta = gamma_prior
    
    for node in tree.traverse("postorder"):
        if not node.is_leaf():
            L[node.name] = np.ones((4,nsites))
            for child in node.children:
                marginal_pt_matrix[child.name] = np.transpose(np.linalg.lstsq(U.T, 
                                np.dot(np.diag(np.power(theta/(theta-D),k)),U.T))[0])
                Down[child.name] = np.dot(marginal_pt_matrix[child.name],L[child.name])
                L[node.name] *= Down[child.name] 
            scaler = np.amax(L[node.name],axis=0)
            L[node.name] /= scaler
            marginalLoglikelihood += np.sum(np.log(scaler))
            
    marginalLoglikelihood += np.sum(np.log(np.dot(pden,L[node.name])))

    return marginalLoglikelihood


from setuptools import setup

setup(name='phyloinfer',
      version='0.4.1',
      description='An efficient Bayesian phylogenetic inference package',
      url='https://github.com/zcrabbit/PhyloInfer',
      author='Cheng Zhang',
      author_email='zc.rabbit@gmail.com',
      license='MIT',
      packages=['phyloinfer'],
      install_requires=[
          'ete3',
          'biopython',
      ],
      zip_safe=False)
